@@ -7,21 +7,21 @@
 
 //! UPB FFI wrapper code for use by Rust Protobuf.
 
-use crate::__internal::{Enum, Private};
+use crate::__internal::{Enum, Private, SealedInternal};
 use crate::{
-    Map, MapIter, MapMut, MapView, Mut, ProtoStr, Proxied, ProxiedInMapValue, ProxiedInRepeated,
-    Repeated, RepeatedMut, RepeatedView, SettableValue, View, ViewProxy,
+    IntoProxied, Map, MapIter, MapMut, MapView, Message, Mut, ProtoBytes, ProtoStr, ProtoString,
+    Proxied, ProxiedInMapValue, ProxiedInRepeated, Repeated, RepeatedMut, RepeatedView, View,
 };
 use core::fmt::Debug;
-use std::alloc::Layout;
-use std::fmt;
 use std::mem::{size_of, ManuallyDrop, MaybeUninit};
-use std::ops::Deref;
 use std::ptr::{self, NonNull};
 use std::slice;
 use std::sync::OnceLock;
 
+#[cfg(bzl)]
 extern crate upb;
+#[cfg(not(bzl))]
+use crate::upb;
 
 // Temporarily 'pub' since a lot of gencode is directly calling any of the ffi
 // fns.
@@ -52,94 +52,29 @@ const UPB_SCRATCH_SPACE_BYTES: usize = 65_536;
 /// view, we can allocate a large block and refer to that when dealing
 /// with readonly access.
 #[repr(C, align(8))] // align to UPB_MALLOC_ALIGN = 8
+#[doc(hidden)]
 pub struct ScratchSpace([u8; UPB_SCRATCH_SPACE_BYTES]);
 impl ScratchSpace {
-    pub fn zeroed_block(_private: Private) -> RawMessage {
+    pub fn zeroed_block() -> RawMessage {
         static ZEROED_BLOCK: ScratchSpace = ScratchSpace([0; UPB_SCRATCH_SPACE_BYTES]);
         NonNull::from(&ZEROED_BLOCK).cast()
     }
 }
 
-/// Serialized Protobuf wire format data.
-///
-/// It's typically produced by `<Message>::serialize()`.
-pub struct SerializedData {
-    data: NonNull<u8>,
-    len: usize,
+#[doc(hidden)]
+pub type SerializedData = upb::OwnedArenaBox<[u8]>;
 
-    // The arena that owns `data`.
-    _arena: Arena,
-}
+impl SealedInternal for SerializedData {}
 
-impl SerializedData {
-    /// Construct `SerializedData` from raw pointers and its owning arena.
-    ///
-    /// # Safety
-    /// - `arena` must be have allocated `data`
-    /// - `data` must be readable for `len` bytes and not mutate while this
-    ///   struct exists
-    pub unsafe fn from_raw_parts(arena: Arena, data: NonNull<u8>, len: usize) -> Self {
-        SerializedData { _arena: arena, data, len }
-    }
-
-    /// Gets a raw slice pointer.
-    pub fn as_ptr(&self) -> *const [u8] {
-        ptr::slice_from_raw_parts(self.data.as_ptr(), self.len)
-    }
-}
-
-impl Deref for SerializedData {
-    type Target = [u8];
-    fn deref(&self) -> &Self::Target {
-        // SAFETY: `data` is valid for `len` bytes as promised by
-        //         the caller of `SerializedData::from_raw_parts`.
-        unsafe { slice::from_raw_parts(self.data.as_ptr(), self.len) }
-    }
-}
-
-impl fmt::Debug for SerializedData {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        fmt::Debug::fmt(self.deref(), f)
-    }
-}
-
-impl SettableValue<[u8]> for SerializedData {
-    fn set_on<'msg>(self, _private: Private, mut mutator: Mut<'msg, [u8]>)
-    where
-        [u8]: 'msg,
-    {
-        mutator.set(self.as_ref())
-    }
-}
-
-// TODO: Investigate replacing this with direct access to UPB bits.
-pub type MessagePresentMutData<'msg, T> = crate::vtable::RawVTableOptionalMutatorData<'msg, T>;
-pub type MessageAbsentMutData<'msg, T> = crate::vtable::RawVTableOptionalMutatorData<'msg, T>;
-pub type BytesPresentMutData<'msg> = crate::vtable::RawVTableOptionalMutatorData<'msg, [u8]>;
-pub type BytesAbsentMutData<'msg> = crate::vtable::RawVTableOptionalMutatorData<'msg, [u8]>;
-pub type InnerBytesMut<'msg> = crate::vtable::RawVTableMutator<'msg, [u8]>;
-pub type InnerPrimitiveMut<'msg, T> = crate::vtable::RawVTableMutator<'msg, T>;
-
-#[derive(Debug)]
-pub struct MessageVTable {
-    pub getter: unsafe extern "C" fn(msg: RawMessage) -> Option<RawMessage>,
-    pub mut_getter: unsafe extern "C" fn(msg: RawMessage, arena: RawArena) -> RawMessage,
-    pub clearer: unsafe extern "C" fn(msg: RawMessage),
-}
-
-impl MessageVTable {
-    pub const fn new(
-        _private: Private,
-        getter: unsafe extern "C" fn(msg: RawMessage) -> Option<RawMessage>,
-        mut_getter: unsafe extern "C" fn(msg: RawMessage, arena: RawArena) -> RawMessage,
-        clearer: unsafe extern "C" fn(msg: RawMessage),
-    ) -> Self {
-        MessageVTable { getter, mut_getter, clearer }
+impl IntoProxied<ProtoBytes> for SerializedData {
+    fn into_proxied(self, _private: Private) -> ProtoBytes {
+        ProtoBytes { inner: InnerProtoString(self) }
     }
 }
 
 /// The raw contents of every generated message.
 #[derive(Debug)]
+#[doc(hidden)]
 pub struct MessageInner {
     pub msg: RawMessage,
     pub arena: Arena,
@@ -171,6 +106,7 @@ pub struct MessageInner {
 ///   cannot be `Clone` but *can* reborrow itself with `.as_mut()`, which
 ///   converts `&'b mut Mut<'a, T>` to `Mut<'b, T>`.
 #[derive(Clone, Copy, Debug)]
+#[doc(hidden)]
 pub struct MutatorMessageRef<'msg> {
     msg: RawMessage,
     arena: &'msg Arena,
@@ -179,15 +115,11 @@ pub struct MutatorMessageRef<'msg> {
 impl<'msg> MutatorMessageRef<'msg> {
     #[doc(hidden)]
     #[allow(clippy::needless_pass_by_ref_mut)] // Sound construction requires mutable access.
-    pub fn new(_private: Private, msg: &'msg mut MessageInner) -> Self {
+    pub fn new(msg: &'msg mut MessageInner) -> Self {
         MutatorMessageRef { msg: msg.msg, arena: &msg.arena }
     }
 
-    pub fn from_parent(
-        _private: Private,
-        parent_msg: MutatorMessageRef<'msg>,
-        message_field_ptr: RawMessage,
-    ) -> Self {
+    pub fn from_parent(parent_msg: MutatorMessageRef<'msg>, message_field_ptr: RawMessage) -> Self {
         MutatorMessageRef { msg: message_field_ptr, arena: parent_msg.arena }
     }
 
@@ -195,35 +127,49 @@ impl<'msg> MutatorMessageRef<'msg> {
         self.msg
     }
 
-    pub fn arena(&self, _private: Private) -> &Arena {
+    pub fn arena(&self) -> &Arena {
         self.arena
     }
+
+    /// # Safety
+    /// - `msg` must be a valid `RawMessage`
+    /// - `arena` must hold the memory for `msg`
+    pub unsafe fn from_raw_parts(msg: RawMessage, arena: &'msg Arena) -> Self {
+        MutatorMessageRef { msg, arena }
+    }
 }
 
-pub fn copy_bytes_in_arena_if_needed_by_runtime<'msg>(
-    msg_ref: MutatorMessageRef<'msg>,
-    val: &'msg [u8],
-) -> &'msg [u8] {
-    copy_bytes_in_arena(msg_ref.arena, val)
+/// Kernel-specific owned `string` and `bytes` field type.
+#[doc(hidden)]
+pub struct InnerProtoString(OwnedArenaBox<[u8]>);
+
+impl InnerProtoString {
+    pub(crate) fn as_bytes(&self) -> &[u8] {
+        &self.0
+    }
+
+    #[doc(hidden)]
+    pub fn into_raw_parts(self) -> (PtrAndLen, Arena) {
+        let (data_ptr, arena) = self.0.into_parts();
+        (unsafe { data_ptr.as_ref().into() }, arena)
+    }
 }
 
-fn copy_bytes_in_arena<'msg>(arena: &'msg Arena, val: &'msg [u8]) -> &'msg [u8] {
-    // SAFETY: the alignment of `[u8]` is less than `UPB_MALLOC_ALIGN`.
-    let new_alloc = unsafe { arena.alloc(Layout::for_value(val)) };
-    debug_assert_eq!(new_alloc.len(), val.len());
-
-    let start: *mut u8 = new_alloc.as_mut_ptr().cast();
-    // SAFETY:
-    // - `new_alloc` is writeable for `val.len()` bytes.
-    // - After the copy, `new_alloc` is initialized for `val.len()` bytes.
-    unsafe {
-        val.as_ptr().copy_to_nonoverlapping(start, val.len());
-        &*(new_alloc as *mut _ as *mut [u8])
+impl From<&[u8]> for InnerProtoString {
+    fn from(val: &[u8]) -> InnerProtoString {
+        let arena = Arena::new();
+        let in_arena_copy = arena.copy_slice_in(val).unwrap();
+        // SAFETY:
+        // - `in_arena_copy` is valid slice that will live for `arena`'s lifetime and
+        //   this is the only reference in the program to it.
+        // - `in_arena_copy` is a pointer into an allocation on `arena`
+        InnerProtoString(unsafe { OwnedArenaBox::new(Into::into(in_arena_copy), arena) })
     }
 }
 
 /// The raw type-erased version of an owned `Repeated`.
 #[derive(Debug)]
+#[doc(hidden)]
 pub struct InnerRepeated {
     raw: RawRepeatedField,
     arena: Arena,
@@ -231,12 +177,27 @@ pub struct InnerRepeated {
 
 impl InnerRepeated {
     pub fn as_mut(&mut self) -> InnerRepeatedMut<'_> {
-        InnerRepeatedMut::new(Private, self.raw, &self.arena)
+        InnerRepeatedMut::new(self.raw, &self.arena)
+    }
+
+    pub fn raw(&self) -> RawRepeatedField {
+        self.raw
+    }
+
+    pub fn arena(&self) -> &Arena {
+        &self.arena
+    }
+
+    /// # Safety
+    /// - `raw` must be a valid `RawRepeatedField`
+    pub unsafe fn from_raw_parts(raw: RawRepeatedField, arena: Arena) -> Self {
+        Self { raw, arena }
     }
 }
 
 /// The raw type-erased pointer version of `RepeatedMut`.
 #[derive(Clone, Copy, Debug)]
+#[doc(hidden)]
 pub struct InnerRepeatedMut<'msg> {
     pub(crate) raw: RawRepeatedField,
     arena: &'msg Arena,
@@ -244,7 +205,7 @@ pub struct InnerRepeatedMut<'msg> {
 
 impl<'msg> InnerRepeatedMut<'msg> {
     #[doc(hidden)]
-    pub fn new(_private: Private, raw: RawRepeatedField, arena: &'msg Arena) -> Self {
+    pub fn new(raw: RawRepeatedField, arena: &'msg Arena) -> Self {
         InnerRepeatedMut { raw, arena }
     }
 }
@@ -252,50 +213,77 @@ impl<'msg> InnerRepeatedMut<'msg> {
 macro_rules! impl_repeated_base {
     ($t:ty, $elem_t:ty, $ufield:ident, $upb_tag:expr) => {
         #[allow(dead_code)]
+        #[inline]
         fn repeated_new(_: Private) -> Repeated<$t> {
             let arena = Arena::new();
-            Repeated::from_inner(InnerRepeated {
-                raw: unsafe { upb_Array_New(arena.raw(), $upb_tag) },
-                arena,
-            })
+            Repeated::from_inner(
+                Private,
+                InnerRepeated { raw: unsafe { upb_Array_New(arena.raw(), $upb_tag) }, arena },
+            )
         }
         #[allow(dead_code)]
         unsafe fn repeated_free(_: Private, _f: &mut Repeated<$t>) {
             // No-op: the memory will be dropped by the arena.
         }
+        #[inline]
         fn repeated_len(f: View<Repeated<$t>>) -> usize {
             unsafe { upb_Array_Size(f.as_raw(Private)) }
         }
-        fn repeated_push(mut f: Mut<Repeated<$t>>, v: View<$t>) {
+        #[inline]
+        fn repeated_push(mut f: Mut<Repeated<$t>>, v: impl IntoProxied<$t>) {
             let arena = f.raw_arena(Private);
             unsafe {
                 assert!(upb_Array_Append(
                     f.as_raw(Private),
-                    <$t as UpbTypeConversions>::to_message_value_copy_if_required(arena, v),
+                    <$t as UpbTypeConversions>::into_message_value_fuse_if_required(
+                        arena,
+                        v.into_proxied(Private)
+                    ),
                     arena,
                 ));
             }
         }
+        #[inline]
         fn repeated_clear(mut f: Mut<Repeated<$t>>) {
             unsafe {
                 upb_Array_Resize(f.as_raw(Private), 0, f.raw_arena(Private));
             }
         }
+        #[inline]
         unsafe fn repeated_get_unchecked(f: View<Repeated<$t>>, i: usize) -> View<$t> {
             unsafe {
                 <$t as UpbTypeConversions>::from_message_value(upb_Array_Get(f.as_raw(Private), i))
             }
         }
-        unsafe fn repeated_set_unchecked(mut f: Mut<Repeated<$t>>, i: usize, v: View<$t>) {
+        #[inline]
+        unsafe fn repeated_set_unchecked(
+            mut f: Mut<Repeated<$t>>,
+            i: usize,
+            v: impl IntoProxied<$t>,
+        ) {
             let arena = f.raw_arena(Private);
             unsafe {
                 upb_Array_Set(
                     f.as_raw(Private),
                     i,
-                    <$t as UpbTypeConversions>::to_message_value_copy_if_required(arena, v.into()),
+                    <$t as UpbTypeConversions>::into_message_value_fuse_if_required(
+                        arena,
+                        v.into_proxied(Private),
+                    ),
                 )
             }
         }
+        #[inline]
+        fn repeated_reserve(mut f: Mut<Repeated<$t>>, additional: usize) {
+            // SAFETY:
+            // - `upb_Array_Reserve` is unsafe but assumed to be sound when called on a
+            //   valid array.
+            unsafe {
+                let arena = f.raw_arena(Private);
+                let size = upb_Array_Size(f.as_raw(Private));
+                assert!(upb_Array_Reserve(f.as_raw(Private), size + additional, arena));
+            }
+        }
     };
 }
 
@@ -332,6 +320,7 @@ macro_rules! impl_repeated_bytes {
             unsafe impl ProxiedInRepeated for $t {
                 impl_repeated_base!($t, PtrAndLen, str_val, $upb_tag);
 
+                #[inline]
                 fn repeated_copy_from(src: View<Repeated<$t>>, mut dest: Mut<Repeated<$t>>) {
                     let len = src.len();
                     // SAFETY:
@@ -356,7 +345,7 @@ macro_rules! impl_repeated_bytes {
                             len
                         );
                         for (src_ptr, dest_ptr) in src_ptrs.iter().zip(dest_ptrs) {
-                            *dest_ptr = copy_bytes_in_arena(&arena, src_ptr.as_ref()).into();
+                            *dest_ptr = arena.copy_slice_in(src_ptr.as_ref()).unwrap().into();
                         }
                     }
                 }
@@ -365,12 +354,18 @@ macro_rules! impl_repeated_bytes {
     }
 }
 
-impl<'msg, T: ?Sized> RepeatedMut<'msg, T> {
+impl<'msg, T> RepeatedMut<'msg, T> {
     // Returns a `RawArena` which is live for at least `'msg`
     #[doc(hidden)]
     pub fn raw_arena(&mut self, _private: Private) -> RawArena {
         self.inner.arena.raw()
     }
+
+    // Returns an `Arena` which is live for at least `'msg`
+    #[doc(hidden)]
+    pub fn arena(&self, _private: Private) -> &'msg Arena {
+        self.inner.arena
+    }
 }
 
 impl_repeated_primitives!(
@@ -384,7 +379,7 @@ impl_repeated_primitives!(
     (u64, u64, uint64_val, upb::CType::UInt64),
 );
 
-impl_repeated_bytes!((ProtoStr, upb::CType::String), ([u8], upb::CType::Bytes),);
+impl_repeated_bytes!((ProtoString, upb::CType::String), (ProtoBytes, upb::CType::Bytes),);
 
 /// Copy the contents of `src` into `dest`.
 ///
@@ -419,11 +414,10 @@ pub unsafe fn repeated_message_copy_from<T: ProxiedInRepeated>(
 
 /// Cast a `RepeatedView<SomeEnum>` to `RepeatedView<i32>`.
 pub fn cast_enum_repeated_view<E: Enum + ProxiedInRepeated>(
-    private: Private,
     repeated: RepeatedView<E>,
 ) -> RepeatedView<i32> {
     // SAFETY: Reading an enum array as an i32 array is sound.
-    unsafe { RepeatedView::from_raw(private, repeated.as_raw(Private)) }
+    unsafe { RepeatedView::from_raw(Private, repeated.as_raw(Private)) }
 }
 
 /// Cast a `RepeatedMut<SomeEnum>` to `RepeatedMut<i32>`.
@@ -431,7 +425,6 @@ pub fn cast_enum_repeated_view<E: Enum + ProxiedInRepeated>(
 /// Writing an unknown value is sound because all enums
 /// are representationally open.
 pub fn cast_enum_repeated_mut<E: Enum + ProxiedInRepeated>(
-    private: Private,
     repeated: RepeatedMut<E>,
 ) -> RepeatedMut<i32> {
     // SAFETY:
@@ -439,27 +432,49 @@ pub fn cast_enum_repeated_mut<E: Enum + ProxiedInRepeated>(
     // - No shared mutation is possible through the output.
     unsafe {
         let InnerRepeatedMut { arena, raw, .. } = repeated.inner;
-        RepeatedMut::from_inner(private, InnerRepeatedMut { arena, raw })
+        RepeatedMut::from_inner(Private, InnerRepeatedMut { arena, raw })
     }
 }
 
+/// Cast a `RepeatedMut<SomeEnum>` to `RepeatedMut<i32>` and call
+/// repeated_reserve.
+pub fn reserve_enum_repeated_mut<E: Enum + ProxiedInRepeated>(
+    repeated: RepeatedMut<E>,
+    additional: usize,
+) {
+    let int_repeated = cast_enum_repeated_mut(repeated);
+    ProxiedInRepeated::repeated_reserve(int_repeated, additional);
+}
+
+pub fn new_enum_repeated<E: Enum + ProxiedInRepeated>() -> Repeated<E> {
+    let arena = Arena::new();
+    // SAFETY:
+    // - `upb_Array_New` is unsafe but assumed to be sound when called on a valid
+    //   arena.
+    unsafe {
+        let raw = upb_Array_New(arena.raw(), upb::CType::Int32);
+        Repeated::from_inner(Private, InnerRepeated::from_raw_parts(raw, arena))
+    }
+}
+
+pub fn free_enum_repeated<E: Enum + ProxiedInRepeated>(_repeated: &mut Repeated<E>) {
+    // No-op: the memory will be dropped by the arena.
+}
+
 /// Returns a static empty RepeatedView.
-pub fn empty_array<T: ?Sized + ProxiedInRepeated>() -> RepeatedView<'static, T> {
+pub fn empty_array<T: ProxiedInRepeated>() -> RepeatedView<'static, T> {
     // TODO: Consider creating a static empty array in C.
 
     // Use `i32` for a shared empty repeated for all repeated types in the program.
-    static EMPTY_REPEATED_VIEW: OnceLock<RepeatedView<'static, i32>> = OnceLock::new();
+    static EMPTY_REPEATED_VIEW: OnceLock<Repeated<i32>> = OnceLock::new();
 
     // SAFETY:
     // - Because the repeated is never mutated, the repeated type is unused and
     //   therefore valid for `T`.
-    // - The view is leaked for `'static`.
     unsafe {
         RepeatedView::from_raw(
             Private,
-            EMPTY_REPEATED_VIEW
-                .get_or_init(|| Box::leak(Box::new(Repeated::new())).as_mut().into_view())
-                .as_raw(Private),
+            EMPTY_REPEATED_VIEW.get_or_init(Repeated::new).as_view().as_raw(Private),
         )
     }
 }
@@ -467,8 +482,8 @@ pub fn empty_array<T: ?Sized + ProxiedInRepeated>() -> RepeatedView<'static, T>
 /// Returns a static empty MapView.
 pub fn empty_map<K, V>() -> MapView<'static, K, V>
 where
-    K: Proxied + ?Sized,
-    V: ProxiedInMapValue<K> + ?Sized,
+    K: Proxied,
+    V: ProxiedInMapValue<K>,
 {
     // TODO: Consider creating a static empty map in C.
 
@@ -483,7 +498,7 @@ where
     //
     // If we used a larger key, then UPB would hash more bytes of the key than Rust
     // initialized.
-    static EMPTY_MAP_VIEW: OnceLock<MapView<'static, bool, bool>> = OnceLock::new();
+    static EMPTY_MAP_VIEW: OnceLock<Map<bool, bool>> = OnceLock::new();
 
     // SAFETY:
     // - The map is empty and never mutated.
@@ -492,14 +507,8 @@ where
     //   The map is empty, therefore it doesn't matter what hash is computed, but we
     //   have to use `bool` type as the smallest key possible (otherwise UPB would
     //   read more bytes than Rust allocated).
-    // - The view is leaked for `'static`.
     unsafe {
-        MapView::from_raw(
-            Private,
-            EMPTY_MAP_VIEW
-                .get_or_init(|| Box::leak(Box::new(Map::new())).as_mut().into_view())
-                .as_raw(Private),
-        )
+        MapView::from_raw(Private, EMPTY_MAP_VIEW.get_or_init(Map::new).as_view().as_raw(Private))
     }
 }
 
@@ -509,16 +518,23 @@ impl<'msg, K: ?Sized, V: ?Sized> MapMut<'msg, K, V> {
     pub fn raw_arena(&mut self, _private: Private) -> RawArena {
         self.inner.arena.raw()
     }
+
+    // Returns an `Arena` which is live for at least `'msg`
+    #[doc(hidden)]
+    pub fn arena(&self, _private: Private) -> &'msg Arena {
+        self.inner.arena
+    }
 }
 
 #[derive(Debug)]
+#[doc(hidden)]
 pub struct InnerMap {
     pub(crate) raw: RawMap,
     arena: Arena,
 }
 
 impl InnerMap {
-    pub fn new(_private: Private, raw: RawMap, arena: Arena) -> Self {
+    pub fn new(raw: RawMap, arena: Arena) -> Self {
         Self { raw, arena }
     }
 
@@ -528,6 +544,7 @@ impl InnerMap {
 }
 
 #[derive(Clone, Copy, Debug)]
+#[doc(hidden)]
 pub struct InnerMapMut<'msg> {
     pub(crate) raw: RawMap,
     arena: &'msg Arena,
@@ -535,36 +552,52 @@ pub struct InnerMapMut<'msg> {
 
 #[doc(hidden)]
 impl<'msg> InnerMapMut<'msg> {
-    pub fn new(_private: Private, raw: RawMap, arena: &'msg Arena) -> Self {
+    pub fn new(raw: RawMap, arena: &'msg Arena) -> Self {
         InnerMapMut { raw, arena }
     }
 
     #[doc(hidden)]
-    pub fn as_raw(&self, _private: Private) -> RawMap {
+    pub fn as_raw(&self) -> RawMap {
         self.raw
     }
 
+    pub fn arena(&mut self) -> &Arena {
+        self.arena
+    }
+
     #[doc(hidden)]
-    pub fn raw_arena(&self, _private: Private) -> RawArena {
+    pub fn raw_arena(&mut self) -> RawArena {
         self.arena.raw()
     }
 }
 
 pub trait UpbTypeConversions: Proxied {
     fn upb_type() -> upb::CType;
+
     fn to_message_value(val: View<'_, Self>) -> upb_MessageValue;
 
     /// # Safety
     /// - `raw_arena` must point to a valid upb arena.
-    unsafe fn to_message_value_copy_if_required(
+    unsafe fn into_message_value_fuse_if_required(
         raw_arena: RawArena,
-        val: View<'_, Self>,
+        val: Self,
     ) -> upb_MessageValue;
 
     /// # Safety
     /// - `msg` must be the correct variant for `Self`.
     /// - `msg` pointers must point to memory valid for `'msg` lifetime.
     unsafe fn from_message_value<'msg>(msg: upb_MessageValue) -> View<'msg, Self>;
+
+    /// # Safety
+    /// - `msg` must be the correct variant for `Self`.
+    /// - `msg` pointers must point to memory valid for `'msg` lifetime.
+    #[allow(unused_variables)]
+    unsafe fn from_message_mut<'msg>(msg: *mut upb_Message, arena: &'msg Arena) -> Mut<'msg, Self>
+    where
+        Self: Message,
+    {
+        panic!("mut_from_message_value is only implemented for messages.")
+    }
 }
 
 macro_rules! impl_upb_type_conversions_for_scalars {
@@ -582,7 +615,7 @@ macro_rules! impl_upb_type_conversions_for_scalars {
                 }
 
                 #[inline(always)]
-                unsafe fn to_message_value_copy_if_required(_: RawArena, val: View<'_, $t>) -> upb_MessageValue {
+                unsafe fn into_message_value_fuse_if_required(_: RawArena, val: $t) -> upb_MessageValue {
                     Self::to_message_value(val)
                 }
 
@@ -605,58 +638,61 @@ impl_upb_type_conversions_for_scalars!(
     bool, bool_val, upb::CType::Bool, false;
 );
 
-impl UpbTypeConversions for [u8] {
+impl UpbTypeConversions for ProtoBytes {
     fn upb_type() -> upb::CType {
         upb::CType::Bytes
     }
 
-    fn to_message_value(val: View<'_, [u8]>) -> upb_MessageValue {
+    fn to_message_value(val: View<'_, ProtoBytes>) -> upb_MessageValue {
         upb_MessageValue { str_val: val.into() }
     }
 
-    unsafe fn to_message_value_copy_if_required(
-        raw_arena: RawArena,
-        val: View<'_, [u8]>,
+    unsafe fn into_message_value_fuse_if_required(
+        raw_parent_arena: RawArena,
+        val: ProtoBytes,
     ) -> upb_MessageValue {
         // SAFETY: The arena memory is not freed due to `ManuallyDrop`.
-        let arena = ManuallyDrop::new(unsafe { Arena::from_raw(raw_arena) });
-        let copied = copy_bytes_in_arena(&arena, val);
-        let msg_val = Self::to_message_value(copied);
-        msg_val
+        let parent_arena = ManuallyDrop::new(unsafe { Arena::from_raw(raw_parent_arena) });
+
+        let (view, arena) = val.inner.into_raw_parts();
+        parent_arena.fuse(&arena);
+
+        upb_MessageValue { str_val: view }
     }
 
-    unsafe fn from_message_value<'msg>(msg: upb_MessageValue) -> View<'msg, [u8]> {
+    unsafe fn from_message_value<'msg>(msg: upb_MessageValue) -> View<'msg, ProtoBytes> {
         unsafe { msg.str_val.as_ref() }
     }
 }
 
-impl UpbTypeConversions for ProtoStr {
+impl UpbTypeConversions for ProtoString {
     fn upb_type() -> upb::CType {
         upb::CType::String
     }
 
-    fn to_message_value(val: View<'_, ProtoStr>) -> upb_MessageValue {
+    fn to_message_value(val: View<'_, ProtoString>) -> upb_MessageValue {
         upb_MessageValue { str_val: val.as_bytes().into() }
     }
 
-    unsafe fn to_message_value_copy_if_required(
+    unsafe fn into_message_value_fuse_if_required(
         raw_arena: RawArena,
-        val: View<'_, ProtoStr>,
+        val: ProtoString,
     ) -> upb_MessageValue {
         // SAFETY: `raw_arena` is valid as promised by the caller
         unsafe {
-            <[u8] as UpbTypeConversions>::to_message_value_copy_if_required(
+            <ProtoBytes as UpbTypeConversions>::into_message_value_fuse_if_required(
                 raw_arena,
-                val.as_bytes(),
+                val.into(),
             )
         }
     }
 
-    unsafe fn from_message_value<'msg>(msg: upb_MessageValue) -> View<'msg, ProtoStr> {
+    unsafe fn from_message_value<'msg>(msg: upb_MessageValue) -> View<'msg, ProtoString> {
         unsafe { ProtoStr::from_utf8_unchecked(msg.str_val.as_ref()) }
     }
 }
 
+#[doc(hidden)]
 pub struct RawMapIter {
     // TODO: Replace this `RawMap` with the const type.
     map: RawMap,
@@ -664,18 +700,14 @@ pub struct RawMapIter {
 }
 
 impl RawMapIter {
-    pub fn new(_private: Private, map: RawMap) -> Self {
-        // SAFETY: __rust_proto_kUpb_Map_Begin is never modified
-        RawMapIter { map, iter: unsafe { __rust_proto_kUpb_Map_Begin } }
+    pub fn new(map: RawMap) -> Self {
+        RawMapIter { map, iter: UPB_MAP_BEGIN }
     }
 
     /// # Safety
     /// - `self.map` must be valid, and remain valid while the return value is
     ///   in use.
-    pub unsafe fn next_unchecked(
-        &mut self,
-        _private: Private,
-    ) -> Option<(upb_MessageValue, upb_MessageValue)> {
+    pub unsafe fn next_unchecked(&mut self) -> Option<(upb_MessageValue, upb_MessageValue)> {
         let mut key = MaybeUninit::uninit();
         let mut value = MaybeUninit::uninit();
         // SAFETY: the `map` is valid as promised by the caller
@@ -685,104 +717,120 @@ impl RawMapIter {
     }
 }
 
-macro_rules! impl_ProxiedInMapValue_for_non_generated_value_types {
-    ($key_t:ty ; $($t:ty),*) => {
-         $(
-            impl ProxiedInMapValue<$key_t> for $t {
-                fn map_new(_private: Private) -> Map<$key_t, Self> {
-                    let arena = Arena::new();
-                    let raw = unsafe {
-                        upb_Map_New(arena.raw(),
-                            <$key_t as UpbTypeConversions>::upb_type(),
-                            <$t as UpbTypeConversions>::upb_type())
-                    };
-                    Map::from_inner(Private, InnerMap { raw, arena })
-                }
+impl<Key, MessageType> ProxiedInMapValue<Key> for MessageType
+where
+    Key: Proxied + UpbTypeConversions,
+    MessageType: Proxied + UpbTypeConversions,
+{
+    fn map_new(_private: Private) -> Map<Key, Self> {
+        let arena = Arena::new();
+        let raw = unsafe {
+            upb_Map_New(
+                arena.raw(),
+                <Key as UpbTypeConversions>::upb_type(),
+                <Self as UpbTypeConversions>::upb_type(),
+            )
+        };
 
-                unsafe fn map_free(_private: Private, _map: &mut Map<$key_t, Self>) {
-                    // No-op: the memory will be dropped by the arena.
-                }
+        Map::from_inner(Private, InnerMap::new(raw, arena))
+    }
 
-                fn map_clear(mut map: Mut<'_, Map<$key_t, Self>>) {
-                    unsafe {
-                        upb_Map_Clear(map.as_raw(Private));
-                    }
-                }
+    unsafe fn map_free(_private: Private, _map: &mut Map<Key, Self>) {
+        // No-op: the memory will be dropped by the arena.
+    }
 
-                fn map_len(map: View<'_, Map<$key_t, Self>>) -> usize {
-                    unsafe {
-                        upb_Map_Size(map.as_raw(Private))
-                    }
-                }
+    fn map_clear(mut map: MapMut<Key, Self>) {
+        unsafe {
+            upb_Map_Clear(map.as_raw(Private));
+        }
+    }
 
-                fn map_insert(mut map: Mut<'_, Map<$key_t, Self>>, key: View<'_, $key_t>, value: View<'_, Self>) -> bool {
-                    let arena = map.raw_arena(Private);
-                    unsafe {
-                        upb_Map_InsertAndReturnIfInserted(
-                            map.as_raw(Private),
-                            <$key_t as UpbTypeConversions>::to_message_value(key),
-                            <$t as UpbTypeConversions>::to_message_value_copy_if_required(arena, value),
-                            arena
-                        )
-                    }
-                }
+    fn map_len(map: MapView<Key, Self>) -> usize {
+        unsafe { upb_Map_Size(map.as_raw(Private)) }
+    }
 
-                fn map_get<'a>(map: View<'a, Map<$key_t, Self>>, key: View<'_, $key_t>) -> Option<View<'a, Self>> {
-                    let mut val = MaybeUninit::uninit();
-                    let found = unsafe {
-                        upb_Map_Get(map.as_raw(Private), <$key_t as UpbTypeConversions>::to_message_value(key),
-                            val.as_mut_ptr())
-                    };
-                    if !found {
-                        return None;
-                    }
-                    Some(unsafe { <$t as UpbTypeConversions>::from_message_value(val.assume_init()) })
-                }
+    fn map_insert(
+        mut map: MapMut<Key, Self>,
+        key: View<'_, Key>,
+        value: impl IntoProxied<Self>,
+    ) -> bool {
+        let arena = map.inner(Private).raw_arena();
+        unsafe {
+            upb_Map_InsertAndReturnIfInserted(
+                map.as_raw(Private),
+                <Key as UpbTypeConversions>::to_message_value(key),
+                <Self as UpbTypeConversions>::into_message_value_fuse_if_required(
+                    arena,
+                    value.into_proxied(Private),
+                ),
+                arena,
+            )
+        }
+    }
 
-                fn map_remove(mut map: Mut<'_, Map<$key_t, Self>>, key: View<'_, $key_t>) -> bool {
-                    unsafe {
-                        upb_Map_Delete(map.as_raw(Private),
-                            <$key_t as UpbTypeConversions>::to_message_value(key),
-                            ptr::null_mut())
-                    }
-                }
+    fn map_get<'a>(map: MapView<'a, Key, Self>, key: View<'_, Key>) -> Option<View<'a, Self>> {
+        let mut val = MaybeUninit::uninit();
+        let found = unsafe {
+            upb_Map_Get(
+                map.as_raw(Private),
+                <Key as UpbTypeConversions>::to_message_value(key),
+                val.as_mut_ptr(),
+            )
+        };
+        if !found {
+            return None;
+        }
+        Some(unsafe { <Self as UpbTypeConversions>::from_message_value(val.assume_init()) })
+    }
 
-                fn map_iter(map: View<'_, Map<$key_t, Self>>) -> MapIter<'_, $key_t, Self> {
-                    // SAFETY: View<Map<'_,..>> guarantees its RawMap outlives '_.
-                    unsafe {
-                        MapIter::from_raw(Private, RawMapIter::new(Private, map.as_raw(Private)))
-                    }
-                }
+    fn map_get_mut<'a>(mut map: MapMut<'a, Key, Self>, key: View<'_, Key>) -> Option<Mut<'a, Self>>
+    where
+        Self: Message,
+    {
+        // SAFETY: The map is valid as promised by the caller.
+        let val = unsafe {
+            upb_Map_GetMutable(
+                map.as_raw(Private),
+                <Key as UpbTypeConversions>::to_message_value(key),
+            )
+        };
+        if val.is_null() {
+            return None;
+        }
+        // SAFETY: The lifetime of the MapMut is guaranteed to outlive the returned Mut.
+        Some(unsafe { <Self as UpbTypeConversions>::from_message_mut(val, map.arena(Private)) })
+    }
 
-                fn map_iter_next<'a>(
-                    iter: &mut MapIter<'a, $key_t, Self>
-                ) -> Option<(View<'a, $key_t>, View<'a, Self>)> {
-                    // SAFETY: MapIter<'a, ..> guarantees its RawMapIter outlives 'a.
-                    unsafe { iter.as_raw_mut(Private).next_unchecked(Private) }
-                        // SAFETY: MapIter<K, V> returns key and values message values
-                        //         with the variants for K and V active.
-                        .map(|(k, v)| unsafe {(
-                            <$key_t as UpbTypeConversions>::from_message_value(k),
-                            <$t as UpbTypeConversions>::from_message_value(v),
-                        )})
-                }
-            }
-         )*
+    fn map_remove(mut map: MapMut<Key, Self>, key: View<'_, Key>) -> bool {
+        unsafe {
+            upb_Map_Delete(
+                map.as_raw(Private),
+                <Key as UpbTypeConversions>::to_message_value(key),
+                ptr::null_mut(),
+            )
+        }
+    }
+    fn map_iter(map: MapView<Key, Self>) -> MapIter<Key, Self> {
+        // SAFETY: MapView<'_,..>> guarantees its RawMap outlives '_.
+        unsafe { MapIter::from_raw(Private, RawMapIter::new(map.as_raw(Private))) }
+    }
+
+    fn map_iter_next<'a>(
+        iter: &mut MapIter<'a, Key, Self>,
+    ) -> Option<(View<'a, Key>, View<'a, Self>)> {
+        // SAFETY: MapIter<'a, ..> guarantees its RawMapIter outlives 'a.
+        unsafe { iter.as_raw_mut(Private).next_unchecked() }
+            // SAFETY: MapIter<K, V> returns key and values message values
+            //         with the variants for K and V active.
+            .map(|(k, v)| unsafe {
+                (
+                    <Key as UpbTypeConversions>::from_message_value(k),
+                    <Self as UpbTypeConversions>::from_message_value(v),
+                )
+            })
     }
 }
 
-macro_rules! impl_ProxiedInMapValue_for_key_types {
-    ($($t:ty),*) => {
-        $(
-            impl_ProxiedInMapValue_for_non_generated_value_types!(
-                $t ; f32, f64, i32, u32, i64, u64, bool, ProtoStr, [u8]
-            );
-        )*
-    }
-}
-
-impl_ProxiedInMapValue_for_key_types!(i32, u32, i64, u64, bool, ProtoStr);
-
 /// `upb_Map_Insert`, but returns a `bool` for whether insert occurred.
 ///
 /// Returns `true` if the entry was newly inserted.
@@ -808,40 +856,3 @@ pub unsafe fn upb_Map_InsertAndReturnIfInserted(
         upb::MapInsertStatus::OutOfMemory => panic!("map arena is out of memory"),
     }
 }
-
-#[cfg(test)]
-mod tests {
-    use super::*;
-    use googletest::prelude::*;
-
-    #[test]
-    fn test_serialized_data_roundtrip() {
-        let arena = Arena::new();
-        let original_data = b"Hello world";
-        let len = original_data.len();
-
-        let serialized_data = unsafe {
-            SerializedData::from_raw_parts(
-                arena,
-                NonNull::new(original_data as *const _ as *mut _).unwrap(),
-                len,
-            )
-        };
-        assert_that!(&*serialized_data, eq(b"Hello world"));
-    }
-
-    #[test]
-    fn assert_c_type_sizes() {
-        // TODO: add these same asserts in C++.
-        use std::ffi::c_void;
-        use std::mem::{align_of, size_of};
-        assert_that!(
-            size_of::<upb_MessageValue>(),
-            eq(size_of::<*const c_void>() + size_of::<usize>())
-        );
-        assert_that!(align_of::<upb_MessageValue>(), eq(align_of::<*const c_void>()));
-
-        assert_that!(size_of::<upb_MutableMessageValue>(), eq(size_of::<*const c_void>()));
-        assert_that!(align_of::<upb_MutableMessageValue>(), eq(align_of::<*const c_void>()));
-    }
-}
